#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "plugin.hpp"
#include "system.hpp"
#include "instrumentation.hpp"
#include "instrumentationcontext_decl.hpp"
#include "Ayudame.h"
#include <cassert>
#include "smpdd.hpp" // FIXME: this the include should not be here (just testing smpdd)

#include <pthread.h>

extern const char*   nanos_get_runtime_version() __attribute__((weak));




#define VERBOSE 0
//#define AYU_NO_THREAD

//#define NDEBUG

namespace nanos {

class InstrumentationAyudame: public Instrumentation {

#ifndef NANOS_INSTRUMENTATION_ENABLED
public:
    // constructor
    InstrumentationAyudame( ) : Instrumentation() {}

    // destructor
    ~InstrumentationAyudame() {}

    // low-level instrumentation interface (mandatory functions)
    void initialize( void ) {}
    void finalize( void ) {}
    void disable( void ) {}
    void enable( void ) {}
    void addEventList ( unsigned int count, Event *events ) {}
    void addResumeTask( WorkDescriptor &w ) {}
    void addSuspendTask( WorkDescriptor &w ) {}
    void threadStart( BaseThread &thread ) {}
    void threadFinish ( BaseThread &thread ) {}
    void incrementMaxThreads( void ) {}
#else


private:
    std::vector < uint64_t > funcid_vec;
    std::vector < uint64_t > threadid_temanejo_threadid_vec;
    std::vector < uint64_t > threadid_vec;
    std::multimap < int, std::pair< int, nanos_event_value_t > > virtuel_task_multimap;


    pthread_mutex_t funcid_vec_mutex;
    pthread_mutex_t threadid_vec_mutex;
    pthread_mutex_t threadid_temanejo_threadid_vec_mutex;
    pthread_mutex_t virtuel_task_multimap_mutex;



    /**
     * translate wd to function Id that is suitable for Ayudame,
     * indicate if function Id not known yet
     *
     * @param wd WorkDescriptor
     * @param funcid is the output of the function
     * @return true if funcid previosuly unknown -> need to register funcid;
     *         false if already known/registered
     */
    bool wd2funcid(WorkDescriptor *wd, int64_t *funcid) {
        uint64_t nanos_id = (uint64_t) ((ext::SMPDD &) (wd->getActiveDevice())).getWorkFct();
        bool need_register = false;

        pthread_mutex_lock(&funcid_vec_mutex);

        // check if nanos_id known already
        if (std::find(funcid_vec.begin(), funcid_vec.end(), nanos_id)==funcid_vec.end()) {
            // unknown funcid
            need_register = true;
            funcid_vec.push_back(nanos_id);
        }

        pthread_mutex_unlock(&funcid_vec_mutex);

        *funcid = (int64_t) nanos_id;
        return need_register;
    }

#ifndef AYU_NO_THREAD

    /**
     * function maps the system thread id to an identifier starting at 1
     * @return
     */
    int64_t get_thread_id() {
        uint64_t thread_id = (uint64_t)  pthread_self();
        int64_t index = 0;

        pthread_mutex_lock(&threadid_temanejo_threadid_vec_mutex);

        std::vector<uint64_t>::iterator it =std::find(threadid_temanejo_threadid_vec.begin(), threadid_temanejo_threadid_vec.end(), thread_id);

        if (it==threadid_temanejo_threadid_vec.end()) {
            threadid_temanejo_threadid_vec.push_back(thread_id);
            index = threadid_temanejo_threadid_vec.size()-1;
        } else {
            index = std::distance(threadid_temanejo_threadid_vec.begin(),it);
        }
        pthread_mutex_unlock(&threadid_temanejo_threadid_vec_mutex);

        return index ;
    }

#endif //AYU_NO_THREAD

    /**
     *
     * @param wd_id_num
     * @param wd
     * @param num_deps
     * @param da
     */
    void addTask(int64_t wd_id_num, WorkDescriptor *wd) {

        int64_t AYU_data[3];
        int64_t funcid;
        WorkDescriptor *parent = wd->getParent();


        if (wd2funcid(wd, &funcid)) { // new function
             char *wd_description = (char*) wd->getDescription();

#if VERBOSE > 0

            fprintf (stderr,
                     "Emitted Event REGISTERFUNCTION for wd %ld, func name: %s\n",
                     wd_id_num, wd_description);
#endif


            AYU_event (AYU_REGISTERFUNCTION, funcid, (void *) wd_description);
        }

        AYU_data[0] = funcid;
        AYU_data[1] = (int64_t) wd->getPriority();     // FIXME: Is Critical/Priority?

        if (parent == NULL) { // not nested, my parent is a thread
            AYU_data[2] = 0;// not nested
        } else { // nested
            AYU_data[2] = parent->getId();// id of parent task
        }
#if VERBOSE > 0
        fprintf (stderr,
                 "Emitting Event ADDTASK with data={%ld,%ld,%ld}, id=%ld\n",
                 AYU_data[0], AYU_data[1], AYU_data[2], wd_id_num);
#endif

        AYU_event (AYU_ADDTASK, wd_id_num, AYU_data);



        /*
                // when the task had no dependence. At submit we are creating the events "create-wd-ptr" and "wd-num-deps".
                // These two events can be used to determines if  (nanos::WD *) (<"create-wd-ptr">)->your_identifier() has dependences or not "<wd-num-deps>" == 0 (ready) and <"wd-num-deps"> != 0 (blocked).

                    for(int i = 0; i< num_deps;i++){
                        if(da[i].isInput()){
                            fprintf (stderr,"dd");
                            return;
                        }
                    }
                    int thread_id = 0;

                    fprintf (stderr,"Emitting AYU_ADDTASKTOQUEUE with data={%ld,%ld,%ld}, id=%ld\n",
                         AYU_data[0], AYU_data[1], AYU_data[2], wd_id_num);
                    AYU_event (AYU_ADDTASKTOQUEUE, (int64_t) wd->getId(), (void*)thread_id);

                    */

    }


    /**
     *
     * @param sender_id
     * @param receiver_id
     * @param dep_direction_value
     * @param dep_address_value
     */
    void addDependency(int sender_id, int receiver_id, nanos_event_value_t dep_direction_value, nanos_event_value_t dep_address_value) {

        int64_t AYU_data[3];
        int64_t  address_id = (int64_t)  dep_address_value;

        switch (dep_direction_value) {

        case 1:     // True dependence

#if VERBOSE > 0

            fprintf (stderr,"++DEP RaW dependence %d -> %d \n",sender_id,receiver_id);
#endif

            pthread_mutex_lock(&threadid_vec_mutex);
            if ( !(std::find(threadid_vec.begin(), threadid_vec.end(), (uint64_t)receiver_id)!=threadid_vec.end()) &&
                    !(std::find(threadid_vec.begin(), threadid_vec.end(), (uint64_t)sender_id)  !=threadid_vec.end()) ) {

                AYU_data[0] = sender_id;
                AYU_data[1] = address_id;
                AYU_data[2] = address_id;
                AYU_event(AYU_ADDDEPENDENCY, (int64_t) receiver_id, AYU_data );
#if VERBOSE > 0

                fprintf (stderr,
                         "Emitting Event ADDDEPENDENCY with data={%ld,%ld,%ld}, %d->%d\n",
                         AYU_data[0], AYU_data[1], AYU_data[2], sender_id, receiver_id );
#endif

            } else {
                //---------------------------------
                //TASK WAIT ON
                //---------------------------------
#if VERBOSE > 0
                fprintf (stderr,"Emitting Event ADDDEPENDENCY to a thread \n" );
#endif

                // add a virtual task to generate an dependency for taskwait on
                AYU_data[0] = -1;
                AYU_data[1] = 0;
                AYU_data[1] = receiver_id;
                AYU_event (AYU_WAITON,  receiver_id, AYU_data);

                // add dependency to the virtual task with an the memory address of the variable

                AYU_data[0] = sender_id;
                AYU_data[1] = address_id;
                AYU_data[2] = address_id;
                AYU_event(AYU_ADDDEPENDENCY, receiver_id, AYU_data );

            }
            pthread_mutex_unlock(&threadid_vec_mutex);


            break;

        case 2:    // Anti-dependence

#if VERBOSE > 0

            fprintf (stderr,"++DEP WaR dependence %d -> %d \n",sender_id,receiver_id);
#endif

            break;

        case 3:     // Output dependence

#if VERBOSE > 0

            fprintf (stderr,"++DEP WaW dependence %d -> %d \n",sender_id,receiver_id);
#endif

            break;

        case 4:     // Output dependence   wd -> concurrent

#if VERBOSE > 0

            fprintf (stderr,"++DEP CloseConcurrent dependence (wd -> concurrent) %d -> %d  data={%ld}\n",sender_id,receiver_id, (int64_t)dep_address_value);
#endif

            pthread_mutex_lock(&virtuel_task_multimap_mutex);
            virtuel_task_multimap.insert(std::make_pair(receiver_id,std::make_pair(sender_id,dep_address_value)));
            pthread_mutex_unlock(&virtuel_task_multimap_mutex);

            break;

        case 5:     // Output dependence concurrent -> wd

#if VERBOSE > 0

            fprintf (stderr,"++DEP OpenConcurrent dependence (concurrent -> wd) %d -> %d \n",sender_id,receiver_id);
#endif

            break;

        case 6:     // wd -> commutative
#if VERBOSE > 0

            fprintf (stderr,"++DEP CloseCommutative  dependence (wd -> commutative) %d -> %d \n",sender_id,receiver_id);
#endif

            pthread_mutex_lock(&virtuel_task_multimap_mutex);
            virtuel_task_multimap.insert(std::make_pair(receiver_id,std::make_pair(sender_id,dep_address_value)));
            pthread_mutex_unlock(&virtuel_task_multimap_mutex);

            break;

        case 7:     // commutative -> wd
#if VERBOSE > 0

            fprintf (stderr,"++DEP OpenCommutative dependence (commutative -> wd) %d -> %d \n",sender_id,receiver_id);
#endif

            break;

        case 8:     // wd -> common
#if VERBOSE > 0

            fprintf (stderr,"++DEP CloseGeneral  dependence (wd -> common)  %d -> %d \n",sender_id,receiver_id);
#endif

            break;

        case 9:     // common -> wd
#if VERBOSE > 0

            fprintf (stderr,"++DEP OpenGeneral dependence   (common -> wd) %d -> %d \n",sender_id,receiver_id);
#endif

            pthread_mutex_lock(&virtuel_task_multimap_mutex);
            for(std::multimap<int,std::pair<int,nanos_event_value_t> > ::const_iterator i(virtuel_task_multimap.begin()),end(virtuel_task_multimap.end()); i!=end; ++i) {
                if(i->first==sender_id) {
                    AYU_data[0] = i->second.first;
                    AYU_data[1] = (int64_t) i->second.second;
                    AYU_data[2] = (int64_t) i->second.second;
                    AYU_event(AYU_ADDDEPENDENCY, (int64_t) receiver_id, AYU_data );
                }
            }
            virtuel_task_multimap.erase(sender_id);
            pthread_mutex_unlock(&virtuel_task_multimap_mutex);

            break;

        default:
            fprintf (stderr,"++DEP Unexpected type dependency");
            break;
        }

    }


    /**
     *
     */
    void addTaskWait() {

        /*
         * int_64_t current_wd = myThread->getCurrentWD()->getId());
         * taskwait(current_wd);
         */
        AYU_event(AYU_BARRIER, 0, 0);
        // starting taskwait
    }

    /**
     * task ready
     * @param wd
     */
    void taskStateReady(WorkDescriptor *wd ) {

        if ( !(std::find(threadid_vec.begin(), threadid_vec.end(), (uint64_t)wd->getId())!=threadid_vec.end()) ) {
            int64_t thread_id = 0;
            //fprintf (stderr,"Emitting taskStateReady task_id=%ld thread_id=%ld\n",
            //         (int64_t) wd->getId(), thread_id);
            AYU_event (AYU_ADDTASKTOQUEUE, (int64_t) wd->getId(), &thread_id);
        }
    }

    /**
     * task blocked
     * @param wd
     */
    void taskStateBlocked(WorkDescriptor *wd ) {
        if ( !(std::find(threadid_vec.begin(), threadid_vec.end(), (uint64_t)wd->getId())!=threadid_vec.end()) ) {
            //int64_t thread_id = 0;
            //fprintf (stderr,"Emitting taskStateBlocked task_id=%ld thread_id=%ld\n",
            //        (int64_t) wd->getId(), thread_id);
            //TODO
        }
    }

public:

    /**
     * constructor
     */
    InstrumentationAyudame() : Instrumentation( *NEW InstrumentationContextDisabled() ) {
        pthread_mutex_init(&funcid_vec_mutex, NULL);
        pthread_mutex_init(&threadid_vec_mutex, NULL);
        pthread_mutex_init(&threadid_temanejo_threadid_vec_mutex, NULL);
        pthread_mutex_init(&virtuel_task_multimap_mutex, NULL);

        //fprintf (stderr," InstrumentationAyudame()\n");
    }

    /**
     * destructor
     */
    ~InstrumentationAyudame() {
        //fprintf (stderr," ~InstrumentationAyudame()\n");
    }

    // low-level instrumentation interface (mandatory functions)

    /**
     *
     */
    void initialize( void ) {
        ayu_runtime_t ayu_rt = AYU_RT_OMPSS;
        AYU_event(AYU_PREINIT, 0, (void *) &ayu_rt);
#if VERBOSE > 0

        fprintf(stderr,"Event PREINIT emitted\n");
#endif

        int system_tasks = 0;


#if HAVE_GETNUMPES

        system_tasks = sys.getNumPEs();
#elif HAVE_GETNUMWORKERS

        if(nanos_get_runtime_version) {
            system_tasks = sys.getSMPPlugin()->getNumWorkers();
        } else {
            system_tasks = 0;
        }
#else
#error "Cannot infer number of PEs/workers"
        system_tasks = 0;
#endif // HAVE_GETNUMPES


        unsigned long n_threads = system_tasks;

#if VERBOSE > 0
        fprintf (stderr, "Event INIT about to be emitted, n_threads %lu\n", n_threads);
#endif

        AYU_event (AYU_INIT, 0, &n_threads);
#if VERBOSE > 0

        fprintf (stderr, "Event INIT emitted, n_threads %lu\n", n_threads);
#endif

    }

    /**
     *
     */
    void finalize( void ) {
        AYU_event(AYU_FINISH, 0, NULL);
#if VERBOSE > 0

        fprintf(stderr,"Event FINISH emitted\n");
#endif

    }



    /**
     *
     */
    void disable( void ) {}

    /**
     *
     */
    void enable( void ) {}


    /**
     *
     * @param w
     */
    void addResumeTask( WorkDescriptor &w ) {
        WorkDescriptor *parent = w.getParent();
        if (parent == NULL) { // task is a thread
#if VERBOSE > 0
            fprintf(stderr,"+++ No parent +++\n");
#endif
            // TODO: treat thread creation!

            pthread_mutex_lock(&threadid_vec_mutex);
            threadid_vec.push_back((int64_t)w.getId());
            pthread_mutex_unlock(&threadid_vec_mutex);

        }
        if (parent != NULL) { // task is a real task
#if VERBOSE > 0
            fprintf(stderr,"+++ parent: %d +++\n", parent->getId());
#endif

            //TODO On which thread runs a nested task?
#ifndef AYU_NO_THREAD

            int64_t thread_id = get_thread_id();
#else

            int64_t thread_id = 0;
#endif //AYU_NO_THREAD

            // TODO: On which thread runs a nested task?
#if VERBOSE > 0

            fprintf(stderr,"Emitting Event PRERUNTASK %d %d\n",(int) w.getId(), thread_id);
#endif

            AYU_event (AYU_PRERUNTASK, (int64_t) w.getId(), &thread_id);

#if VERBOSE > 0

            fprintf(stderr,"Emitting Event RUNTASK %d\n",(int) w.getId());
#endif

            AYU_event (AYU_RUNTASK, (int64_t) w.getId(), NULL);
        }
    }


    /**
     *
     * @param w
     * @param last
     */
    void addSuspendTask( WorkDescriptor &w, bool last ) {
        if (last) {
#if VERBOSE > 0
            fprintf(stderr,"Emitting Event REMOVETASK %d\n",(int) w.getId());
#endif

            AYU_event(AYU_REMOVETASK, (int64_t) w.getId(), NULL);
        }else{
#if VERBOSE > 0
            fprintf(stderr,"Emitting Event REMOVETASK NOT LAST ######## %d\n",(int) w.getId());
#endif
        }
    }



    /**
     *
     * @param count
     * @param events
     */
    void addEventList ( unsigned int count, Event *events ) {
        InstrumentationDictionary *iD = sys.getInstrumentation()->getInstrumentationDictionary();
        static const nanos_event_key_t create_wd_id = iD->getEventKey("create-wd-id"); //13
        static const nanos_event_key_t create_wd_ptr = iD->getEventKey("create-wd-ptr");//14
        static const nanos_event_key_t wd_num_deps = iD->getEventKey("wd-num-deps");// 15
        static const nanos_event_key_t wd_deps_ptr = iD->getEventKey("wd-deps-ptr");// 16
        static const nanos_event_key_t dependence = iD->getEventKey("dependence"); //36
        static const nanos_event_key_t funct_location = iD->getEventKey("user-funct-location");//21
        static const nanos_event_key_t user_code  = iD->getEventKey("user-code");
        static const nanos_event_key_t dep_direction = iD->getEventKey("dep-direction");//37
        static const nanos_event_key_t dep_address = iD->getEventKey("dep-address"); //43
        //static const nanos_event_key_t graph_size = iD->getEventKey("graph-size");//23
        static const nanos_event_key_t wd_ready = iD->getEventKey("wd-ready");
        static const nanos_event_key_t wd_blocked = iD->getEventKey("wd-blocked");

        static const nanos_event_key_t api          = getInstrumentationDictionary()->getEventKey("api");
        static const nanos_event_value_t wait_group = getInstrumentationDictionary()->getEventValue("api","wg_wait_completion");

        // 20 22 31 32 18 33 34 not interresting

#if VERBOSE > 2

        fprintf(stderr,"addEventList\n");
        for (unsigned int ii=0; ii<count;ii++) {
            fprintf (stderr,
                     "key=%d value=%d \n",events[ii].getKey(),events[ii].getValue());
        }
#endif

        for (unsigned int i = 0; i < count; i++) {

            int64_t value = events[i].getValue();

            //printf("\t\tKEY:%d \n", events[i].getKey());
            switch ( events[i].getType() ) {
            case NANOS_STATE_START:
                break;
            case NANOS_STATE_END:
                break;
            case NANOS_SUBSTATE_START:
            case NANOS_SUBSTATE_END:
            case NANOS_PTP_START:
            case NANOS_PTP_END:
                break;
            case NANOS_POINT:
                //printf(" NANOS_POINT: key:%d \n", events[i].getKey());
                if ( events[i].getKey() == create_wd_ptr ) { //It's a create task event
                    //----------------read all data to event -----------------
                    WorkDescriptor *wd = (WorkDescriptor *) events[i].getValue();
                    int64_t wd_id_num = wd->getId();

                    //----------------read all data to event -----------------
                    addTask(wd_id_num, wd);
                }

                if ( events[i].getKey() == dependence ) { //It's a dependence event
                    //----------------read all data to event -----------------
                    nanos_event_value_t dependence_value = events[i].getValue();
                    int sender_id = (int) ( dependence_value >> 32 );
                    int receiver_id = (int) ( dependence_value & 0xFFFFFFFFFF );
                    i++;
                    assert(i < count);
                    assert(events[i].getKey() == dep_direction);
                    nanos_event_value_t dep_direction_value = events[i].getValue();
                    i++;
                    assert(i < count);
                    assert(events[i].getKey() == dep_address);
                    nanos_event_value_t dep_address_value = events[i].getValue();
                    //----------------read all data to event -----------------

                    addDependency(sender_id, receiver_id, dep_direction_value, dep_address_value);
                }

                if(events[i].getKey() == wd_ready) {
                    WorkDescriptor *wd = (WorkDescriptor *) events[i].getValue();
                    taskStateReady(wd);
                }

                if(events[i].getKey() == wd_blocked) {
                    WorkDescriptor *wd = (WorkDescriptor *) events[i].getValue();
                    taskStateBlocked(wd);
                }

                break;

            case NANOS_BURST_START:

                if ( events[i].getKey() == user_code ) {
#if VERBOSE > 0
                    fprintf(stderr,"Emitting Event RUNTASK entering usercode %d\n",(int) value);
#endif
                    AYU_event (AYU_RUNTASK, (int64_t) value, NULL);
                }

                // TODO: get function location when task is created!
                // and communicate to ayudame
                if ( events[i].getKey() == funct_location ) {
                    std::string description = iD->getValueDescription( events[i].getKey(), events[i].getValue() );
#if VERBOSE > 0

                    //printf("funct_location\n");
                    //fprintf(stderr,"NANOS++: Executing %s function location (key = %ld, value =%ld)\n",
                    //        description.c_str(),(int64_t)events[i].getKey(),(int64_t)events[i].getValue() );
#endif

                }

                if (  events[i].getKey() == api &&  events[i].getValue() == wait_group )  {
#if VERBOSE > 0
                    fprintf(stderr,"Wait_on key=%ld, value =%ld \n",(int64_t) events[i].getKey(), (int64_t) events[i].getValue()  );
#endif
                    addTaskWait();
                }
                break;
            case NANOS_BURST_END:

                if ( events[i].getKey() == user_code ) {
#if VERBOSE > 0
            fprintf(stderr,"Emitting Event REMOVETASK leaving usercode %d\n",(int) value);
#endif
            AYU_event(AYU_REMOVETASK, (int64_t) value, NULL);
                }
            default:
                break;
            } // switch event type
        } // for loop
#if VERBOSE > 2
        fprintf(stderr,"addEventList done\n\n");
#endif

    } // func: addEventList()

    /**
     *
     * @param thread
     */
void threadStart( BaseThread &thread ) {}

    /**
     *
     * @param thread
     */
    void threadFinish ( BaseThread &thread ) {}

#endif //NANOX_EXTRAE_SUPPORTED_VERSION
}
;

namespace ext {

class InstrumentationAyudamePlugin: public Plugin {
public:
    InstrumentationAyudamePlugin() :
            Plugin(
                "Instrumentation which implements Ayudame/Temanejo protocol.",
        1) {}
    ~InstrumentationAyudamePlugin() {}

    void config(Config &cfg) {}

    void init() {
        sys.setInstrumentation(new InstrumentationAyudame());
    }
};

} // namespace ext

} // namespace nanos

DECLARE_PLUGIN("intrumentation-ayudame",
               nanos::ext::InstrumentationAyudamePlugin)
;
