/****h* Ayu/Ayudame
 *  NAME
 *    Ayudame
 *    - library to communicate between instrumented task parallel runtime
 *    and Temanejo
 *  DESCRIPTION
 *    Ayudame provides a callback function AYU_event() with which task
 *    parallel runtimes can be instrumentd. It processes the information given
 *    to it via the function parameters and passes it to clients via sockets.
 *
 *    |html <img src="Ayudame_scheme.png" alt="Ayudame_scheme.png missing"/>
 *    |latex \includegraphics{Ayudame_scheme.png}
 *
 *    AYU_event() is declared with __attribute__ ((weak)) in Ayudame.h
 *    which has to be included in the runtime.
 *  USAGE
 *    Please refer to the documentation provided with the source for in depth
 *    information on how to install and use Ayudame.
 *
 *    In few words, you have to:
 *    * Instrument your application or runtime with calls to AYU_event or use
 *      a runtime which is already instrumented
 *    * Build Ayudame with the usual "configure-make-make install" procedure
 *    * run your application with libayudame.so preloaded or by passing its
 *      location to Temanejo
 *  SIDE EFFECTS
 *    Some ouput is printed to stdout. For quiet operation reroute
 *    to file or /dev/null or set environment variable AYU_VERBOSE = -1
 *  AUTHOR
 *    Steffen Brinkmann et al, HLRS <temanejo@hlrs.de>
 *  COPYRIGHT
 *    (C) 2010-2015 HLRS, University of Stuttgart
 *    Ayudame is published under the terms of the BSD license.
 * 
 *    The code in the function AYU_timestamp is orginally by 
 *      jbenet taken from https://gist.github.com/jbenet/1087739/
 ******/

#include "Ayudame.h"
#include "Ayudame_lib.h"

static int AYU_client = -1;   // client handle

// mutex for variables AYU_client and for messages
static pthread_mutex_t AYU_clientMutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t AYU_msgMutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t AYU_countMutex = PTHREAD_MUTEX_INITIALIZER;

unsigned long AYU_n_threads;       // current number of threads
unsigned long AYU_max_threads;     // maximum number of threads


/****f* Ayu/AYU_event
 *  NAME
 *    AYU_event
 *    - callback for emitting events
 *  SYNOPSIS
 *    AYU_event(event, taskId, pointer)
 *  DESCRIPTION
 *    this function can be called from a task parallel runtime to communicate
 *    with the Ayudame package. Events of type ayu_event_t, the task ID
 *    of the calling task and a pointer to arbitrary data can be sent.
 *    AYU_event will connect to a client via tcp sockets and pass on
 *    the information.
 *  PARAMETERS
 *    ayu_event_t event      // type of event
 *    const int64_t taskId   // taskId from which the event is sent
 *    void *p                // pointer to further data
 *  RETURN VALUE
 *    void
 *  SOURCE
 ma */
void
AYU_event (ayu_event_t event, const int64_t taskId, void *p)
{
  //int i = 0;      // iteration variable
  int n = -1;     // return of read and write
  int c = -1;     // client handle
  int64_t buf[AYU_buf_size];  // send and recieve buffer
  char string_buf[AYU_string_buf_size]; // string buffer
  //static int preselect_count = 0; // preselect task event count
  static uint64_t regfunc_count = 0;  // unique function id
  static ayu_runtime_t runtimeId = AYU_RT_UNKNOWN;
  static int ayu_event_count = 0; // preselect task event count

  //  return;
  pthread_mutex_lock (&AYU_countMutex);
  ++ayu_event_count;
  pthread_mutex_unlock (&AYU_countMutex);

  uint64_t ts = AYU_timestamp (); // timestamp [nanoseconds]

  std::string bt_file;

  buf[0] = htobe64 (runtimeId);
  buf[1] = htobe64 (taskId);
  buf[2] = htobe64 (event);
  buf[3] = htobe64 (0);
  buf[4] = htobe64 (0);
  buf[5] = htobe64 (0);
  buf[6] = htobe64 (0);
  buf[7] = htobe64 (ts);

  // Identification --------------
  if (is_verbose_greater_than(2)){
    if (event != AYU_PRESELECTTASK)
      {
        printf("%d POSIXID %x\n",ayu_event_count,pthread_self ());
      }
  }
  // return;
  // -----------------------------

  // initialise ---------
  if (event == AYU_PREINIT)
    {
      if (is_verbose_greater_than(1)){
        printf("Ayudame: initialising. version %s\n",AYU_VERSION);
        if (AYU_CONNECTTOSOCKET){
          printf("Ayudame: will connect to socket\n");
        }
        if (AYU_WRITETOSTDOUT){
          printf("Ayudame: will write to stdout\n");
        }
      }
    if ( p == NULL ){
    runtimeId = AYU_RT_SMPSS ;
    }
    else{
      runtimeId = *(ayu_runtime_t *) p;
    }

#if AYU_CONNECTTOSOCKET
      c = AYU_initSocket ();
      if (c < 0)
  {
    fprintf(stderr,"Ayudame: ERROR: AYU_initSocket() failed.\n");
    return;
  }

      // locking for future purpose of disconnect and reconnect to
      // (several) client(s)
      pthread_mutex_lock (&AYU_clientMutex);
      c = AYU_client;
      pthread_mutex_unlock (&AYU_clientMutex);

      buf[0] = htobe64 (runtimeId);
      buf[1] = htobe64 (getpid());

      AYU_write (c, (void *) buf, sizeof (buf));

      AYU_initControl (c);

#endif

      return;
    }
  // --------------------

#if AYU_CONNECTTOSOCKET
  // get client handle -------
  pthread_mutex_lock (&AYU_clientMutex);
  c = AYU_client;
  pthread_mutex_unlock (&AYU_clientMutex);
  // --------------------
#endif

 /*
  // wait on initialisation -------
  for (int i = 0; i < 1000; ++i)
    {
      pthread_mutex_lock (&AYU_clientMutex);
      c = AYU_client;
      pthread_mutex_unlock (&AYU_clientMutex);

      if ( c > 0 ) {break;}
      if ( runtimeId != AYU_RT_UNKNOWN ) {break;}

      fprintf(stderr, "Ayudame: waiting for init.\n");
      usleep (1.e3);
    }
  */

#if AYU_CONNECTTOSOCKET
  if (c < 0)
    {
      fprintf(stderr, "Ayudame: ERROR: no client handle.\n");
      return;
    }
#endif

/*
  if (runtimeId == AYU_RT_UNKNOWN)
    {
      fprintf(stderr, "Ayudame: ERROR: runtime unknown\n");
      return;
    }
*/

  // set the runtime id after making sure, that PREINIT happened
  buf[0] = htobe64 (runtimeId);
  // --------------------

  // catch AYU_INIT ----
  if (event == AYU_INIT){
  if ( p == NULL ){ // for backward compatibility
    AYU_n_threads = 0;
  }
  else {
    AYU_n_threads = *(unsigned long*)p;
  }
    if (AYU_n_threads == 0){
      char *ayu_n_threads_str;
      ayu_n_threads_str = getenv ("CSS_NUM_CPUS");
      if (ayu_n_threads_str != NULL){
        AYU_n_threads = atoi (ayu_n_threads_str);
        if (is_verbose_greater_than(1)){
          printf("Ayudame: found CSS_NUM_CPUS = %ld\n",AYU_n_threads);
        }
      }
      else {
  fprintf(stderr,"Ayudame: number of threads could not be set\n");
      }
    }
    else {
      if (is_verbose_greater_than(1)){
        printf("Ayudame: number of CPUs from runtime = %ld\n",AYU_n_threads);
      }
    }
    AYU_max_threads = AYU_n_threads;
  }
  // --------------------

  // catch AYU_PRESELECTTASK ----
  if (event == AYU_PRESELECTTASK)
    {
      if (is_verbose_greater_than(0)){
        fprintf(stderr,"Ayudame: WARNING: Ignored event AYU_PRESELECTTASK\n");
      }
      return;
    }
  // --------------------

  // catch AYU_REGISTERFUNCTION -------

  if (event == AYU_REGISTERFUNCTION)
    {

      char *funcname;
      uint64_t funcid;

      if ((runtimeId==AYU_RT_SMPSS) && (taskId==0)) {  // SMPSs sends only name of function
  funcname = (char *)p;
  funcid = regfunc_count;
  //regfunc_count++;   // RACE CONDITION HERE
  __sync_add_and_fetch (&regfunc_count, 1);
      } else {                       // others send function id (as taskId) and name of function
  funcname = (char *)p;
  funcid = (uint64_t) taskId;
      }

      size_t string_len = strlen(funcname);
      buf[3] = htobe64 (string_len);
      buf[4] = htobe64 (funcid);

      memset (string_buf, '\0', AYU_string_buf_size);
      strcpy (string_buf, funcname);

      char full_buf[sizeof(buf) + sizeof(string_buf)];
      memcpy(full_buf, buf, sizeof(buf));
      memcpy(full_buf+sizeof(buf), string_buf, string_len);

      n = AYU_write (c, (void *) full_buf, sizeof(buf) + string_len);
      if (n < 0){
        return;
      }

      return;
    }
  // -------------------------------------------------------

  // write to client buffer and send message ---------------

  // -- AYU_INIT -- //
  if (event == AYU_INIT){
    buf[3] = htobe64 (AYU_n_threads);
  }

  // -- AYU_ADDTASK -- //
  else if (event == AYU_ADDTASK){
    int64_t *Ayu_data = (int64_t *) p;
    buf[3] = htobe64 (Ayu_data[0]); // tasks function ID (-1 for hidden tasks)
    buf[4] = htobe64 (Ayu_data[1]); // pass whether critical

    // backtrace ------------------

#if AYU_BACKTRACE == 1
    void *bt_buffer[1000];
    char **bt_strings;
    int bt_nptrs = backtrace(bt_buffer, 1000); // number of addresses
    bt_strings = backtrace_symbols(bt_buffer, bt_nptrs);

    // depth 2 of calltree:
    // 0 is AYU_event,
    // 1 is runtime call function,
    // 2 is actual taskified function
    std::string bt_s(bt_strings[2]);

    std::string bt_prog = bt_s.substr(0,bt_s.find('('));
    std::string bt_addr = bt_s.substr(bt_s.find('[')+1,8);

    // get file and line with addr2line
    std::string bt_exec("addr2line -e " + bt_prog + " -f -p " + bt_addr);
    FILE *bt_out = popen(bt_exec.c_str(),"r");
    char bt_buf2[1000];
    fgets(bt_buf2, 1000, bt_out);
    pclose(bt_out);

    strtok(bt_buf2,"\n"); // strips trailing \n . is it thread safe???

    bt_file = bt_buf2;
    size_t tmp = bt_file.find(':');
    size_t tmp2 = bt_file.find_last_of(' ')+1;

    // This gets the calling function:
    std::string bt_func = bt_file.substr(0,bt_file.find(' '));

    // This gets the line number:
    std::string bt_line_no = bt_file.substr(tmp+1);

    // This gets the file name:
    bt_file = bt_file.substr(tmp2,tmp-tmp2);

    // To get the source code line:
    //    bt_exec = "sed '" + bt_line_no + "!d' " + bt_file; // print one line of file
    //    bt_out = popen(bt_exec.c_str(),"r");
    //    fgets(bt_buf2, 1000, bt_out);
    //    pclose(bt_out);
    //    strtok(bt_buf2,"\n"); // strips trailing \n . is it thread safe???


    //Backtrace_symbols_fd(buffer, nptrs, STDOUT_FILENO);

    //      show_backtrace();

    memset (string_buf, '\0', AYU_string_buf_size); // file name
    strcpy (string_buf, bt_file.c_str());

    buf[5] = htobe64 (bt_file.size()); // length of file name
    buf[6] = htobe64 (atol(bt_line_no.c_str())); // line number

    char full_buf[sizeof(buf) + sizeof(string_buf)];
    memcpy(full_buf, buf, sizeof(buf));
    memcpy(full_buf+sizeof(buf), string_buf, string_len);

#endif // AYU_BACKTRACE

  }

  // -- AYU_ADDDEPENDENCY -- //
  else if (event == AYU_ADDDEPENDENCY) {
    uintptr_t *Ayu_data = (uintptr_t *) p;
    buf[3] = htobe64 (Ayu_data[0]); // pass task ID of last writer
    buf[4] = htobe64 (Ayu_data[1]); // dependency address
    buf[5] = htobe64 (Ayu_data[2]); // original dependency address
  }

  // -- AYU_PRERUNTASK -- //
  else if (event == AYU_PRERUNTASK) {
    buf[3] = htobe64 (*(int *) p);  // thread Id
  }

  // -- AYU_ADDTASKTOQUEUE -- //
  else if (event == AYU_ADDTASKTOQUEUE) {
    buf[3] = htobe64 (*(int *) p);  // thread Id
  }

#if !(AYU_BACKTRACE == 1)          /* default case */
  n = AYU_write (c, (void *) buf, sizeof (buf));
#else  /* AYU_BACKTRACE == 1 */
  if (event == AYU_ADDTASK) {
    n = AYU_write (c, (void *) full_buf, sizeof(buf) + bt_file.size());   // get rid of bt_file.size()
  } else {
    n = AYU_write (c, (void *) buf, sizeof (buf));
  }
#endif // BACKTRACE
  if (n < 0){
    return;
  }

  /*
#if AYU_WRITETOSTDOUT
  printf("Ayudame_EVENT: ");
  for (size_t i=0; i<AYU_buf_size; ++i){
    printf(" %d",be64toh(buf[i]));
  }
  printf("\n");
#if AYU_BACKTRACE == 1
  printf("%s %d\n",string_buf,bt_file.size());;

#endif // BACKTRACE
#endif // AYU_WRITETOSTDOUT
  */
  // --------------------

#if AYU_CONNECTTOSOCKET

  if (event == AYU_INIT ||
      event == AYU_PRERUNTASK ||
      event == AYU_REMOVETASK ||
      event == AYU_WAITON ||
      event == AYU_BARRIER)
    {
      AYU_pause (event, taskId);
    }

#endif

  return;

}

/******/

/****f* Ayu/AYU_initSocket
 *  NAME
 *    AYU_initSocket
 *    - initialise server side communication for the Ayudame tool
 *  SYNOPSIS
 *    AYU_initSocket()
 *  DESCRIPTION
 *    Initialises the server side communication for the Ayudame tool
 *    performing the following tasks:
 *    * check for the environment variable AYU_PORT or generate one
 *    * create a socket
 *    * bind to the port AYU_PORT
 *    * listen for a client to connect
 *    * accept the client
 *    * set AYU_client to the client handle
 *  PARAMETERS
 *    void
 *  RETURN VALUE
 *    returns 0 on success and -1 on failure
 *  TODO
 *    * possibility for various clients to connect
 *  SOURCE
 */

int
AYU_initSocket (void)
{
  // declare variables -----------------------

  int server_socket;    // server socket handle
  int client_socket;    // client socket handle
  int cli_size;     // client socket size
  struct sockaddr_in srv, cli;  // address structs for srv and cli sockets
  unsigned short int ayu_socket_port; // port to which the socket will
  // connect

  // -----------------------------------------

  // initialise variables --------------------

  server_socket = client_socket = cli_size = -1;
  memset (&srv, 0, sizeof (struct sockaddr_in));
  memset (&cli, 0, sizeof (struct sockaddr_in));
  ayu_socket_port = 0;

  // -----------------------------------------

  // read AYU_PORT from environment ---
  char *ayu_socket_port_str;
  ayu_socket_port_str = getenv ("AYU_PORT");
  if (ayu_socket_port_str != NULL)
    {
      ayu_socket_port = atoi (ayu_socket_port_str);
      if (is_verbose_greater_than(1)){
        printf("Ayudame: found AYU_PORT = %d\n",ayu_socket_port);
      }
    }
  else
    {
      ayu_socket_port = getpid () % 1000 + 5000;
      if (is_verbose_greater_than(1)){
        printf("Ayudame: AYU_PORT set to %d\n",ayu_socket_port);
      }
    }
  // -----------------------------------------

  // establish socket ------------------------
  // TODO: Do we need IPv6 support?
  server_socket = socket (AF_INET, SOCK_STREAM, 0);
  if (server_socket == -1)
    {
      perror ("Socket creation failed");
      return -1;
    }

  {
    int optval;

    // set SO_REUSEADDR on a socket to true (1) in order to allow reuse
    optval = 1;
    setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, \
         &optval, sizeof optval);
  }

  srv.sin_addr.s_addr = INADDR_ANY;
  srv.sin_port = htons (ayu_socket_port);
  srv.sin_family = AF_INET;

  if (bind (server_socket, (struct sockaddr *) &srv, sizeof (srv)) == -1)
    {
      char error_msg[128];
      snprintf (error_msg, 127,
    "Bind socket to ayudame port %u failed.", ayu_socket_port);
      perror ((const char *) error_msg);
      return -1;
    }
  // -----------------------------------------

  // listen and accept client -----------------------
  if (listen (server_socket, 1) == -1)
    {
      perror ("listen() failed");
      return -1;
    }

  cli_size = sizeof (cli);
  client_socket =
    accept (server_socket, (struct sockaddr *) &cli,
      (socklen_t *) & cli_size);

  if (client_socket == -1)
    {
      perror ("accept() failed");
      return -1;
    }
    if (is_verbose_greater_than(1)){
      printf("Ayudame: client accepted from %s on port %d\n",inet_ntoa (cli.sin_addr),ayu_socket_port);
    }

  // -----------------------------------------

  pthread_mutex_lock (&AYU_clientMutex);
  AYU_client = client_socket;
  pthread_mutex_unlock (&AYU_clientMutex);

  return 0;
}

/******/

/****f* Ayu/AYU_write
 *  NAME
 *    AYU_write
 *    - write data to a socket connection
 *  SYNOPSIS
 *    n = AYU_write(c,buf,AYU_buf_size)
 *  DESCRIPTION
 *    writes data to a client accessed by a handle
 *  PARAMETERS
 *    int c                // client handle
 *    void *buf            // pointer to data chunk
 *    size_t buf_size      // size of data chunk in bytes
 *  RETURN VALUE
 *    integer number of sent bytes, -1 on failure
 *  SOURCE
 */
int
AYU_write (int c, void *buf, size_t buf_size)
{
#if AYU_CONNECTTOSOCKET
  int n = -1;
  for (unsigned int i = 0; i < buf_size; i += n)
    {
      pthread_mutex_lock (&AYU_msgMutex);
      n = write (c, (uint8_t *) buf + i, buf_size - i);
      pthread_mutex_unlock (&AYU_msgMutex);
      if (n < 0)
    {
      perror ("ERROR writing to client socket");
      sleep (10);
      return -1;
    }
    }
#endif

#if AYU_WRITETOSTDOUT
  char outbuf[AYU_string_buf_size*2];
  char tmpbuf[AYU_string_buf_size*2];
  char sentinel[] = "AYUDAME_SEND: ";

  memset (outbuf, '\0', sizeof(outbuf));
  memset (tmpbuf, '\0', sizeof(outbuf));

  snprintf(outbuf,sizeof(outbuf)-strlen(outbuf), sentinel);  // outbuf = "AYU_SEND: "

  for (unsigned int i = 0; i < AYU_buf_size; ++i) {
    sprintf(tmpbuf," %ld",be64toh (((uint64_t *) buf)[i]));
    strcat(outbuf,tmpbuf);                                  // outbuf = "AYU_SEND: X0 X1 X2 X3 X4 X5 X6 X7"
  }

  if (buf_size>AYU_buf_size*sizeof(uint64_t)) {             // a string is attached to message
    strcpy(tmpbuf, outbuf);
    buf = ((char *)buf) + AYU_buf_size*sizeof(uint64_t);    // ignore first part of buf
    snprintf(outbuf, sizeof(outbuf), "%s %s", tmpbuf, (char*) buf);  // outbuf = "AYU_SEND: X0 X1 X2 X3 X4 X5 X6 X7  MYSTRINGHERE"
  }
  pthread_mutex_lock (&AYU_msgMutex);
  printf("%s\n",outbuf);
  pthread_mutex_unlock (&AYU_msgMutex);
#endif

#if AYU_CONNECTTOSOCKET
  return n;
#else
  return 0;
#endif
}

/******/

/****f* Ayu/AYU_timestamp
 *  NAME
 *    AYU_timestamp
 *    - get timestamp in nanoseconds since Epoch
 *  SYNOPSIS
 *    t = AYU_timestamp()
 *  DESCRIPTION
 *    return timestamp in nanoseconds since Epoch
 *  PARAMETERS
 *    void
 *  RETURN VALUE
 *    integer number of nanoseconds
 *  SOURCE
 */
extern "C"
{
  __inline__ uint64_t AYU_timestamp (void)
  {
    struct timespec ts;
    uint64_t nanos; 

#ifdef __MACH__ // OS X does not have clock_gettime, use clock_get_time
    clock_serv_t cclock;
    mach_timespec_t mts;
    host_get_clock_service(mach_host_self(), CALENDAR_CLOCK, &cclock);
    clock_get_time(cclock, &mts);
    mach_port_deallocate(mach_task_self(), cclock);
    ts.tv_sec = mts.tv_sec;
    ts.tv_nsec = mts.tv_nsec;
#else
    clock_gettime(CLOCK_REALTIME, &ts);
#endif
    
    nanos = ts.tv_nsec + 1000000000L*ts.tv_sec;
    return nanos;
  }
}
/******/

/****f* Ayu/AYU_registerTask
 *  NAME
 *    AYU_registerTask
 *    - obsolete. Kept for backward compatibility
 *  SYNOPSIS
 *
 *  DESCRIPTION
 *
 *  PARAMETERS
 *
 *  RETURN VALUE
 *
 *  SOURCE
 */
extern "C"
{
  void AYU_registerTask(void*)
  {
    fprintf(stderr,"AYU_registerTask is deprecated. Please delete calls.\n");
  }

}
/******/

/****f* Ayu/AYU_is_verbose_greater_than
 *  NAME
 *    AYU_is_verbose_greater_than
 *    - checks whether the verbosity level is greater than a given integer.
 *  SYNOPSIS
 *    if (AYU_is_verbose_greater_than(1)) {printf("debug message\n");}
 *  DESCRIPTION
 *    This function checks whether the environment variable AYU_VERBOSE is set
 *    If it is, it checks whether it is greater than the number given in the
 *    argument, otherwise it checks the argument against a default value (1).
 *  PARAMETERS
 *    int verbosity_level
 *  RETURN VALUE
 *    bool whether AYU_VERBOSE (or default value) is greater than verbosity_level
 *  SOURCE
 */
bool is_verbose_greater_than(int verbosity_level)
{
  int default_verbosity_level = 0;
  if ( NULL != getenv("AYU_VERBOSE") ) {
    return atoi(getenv("AYU_VERBOSE")) > verbosity_level;
  }
  return default_verbosity_level > verbosity_level;
}
