/****H* Ayu/Ayudame.h
 *  NAME
 *    Ayudame.h - header file containing the event and request types of the
 *    Ayudame package and callback function declarations.
 *  DESCRIPTION
 *    This header file declares the following enumeration types:
 *    * ayu_event_t
 *    * ayu_request_t
 *    * ayu_runtime_t
 *    And these function declarations:
 *    * AYU_event
 *    * AYU_getBreakpoint
 *    * AYU_getNumThreads
 *    * AYU_getPriorityLevel
 *    * AYU_isBlocked
 *    * AYU_registerTask
 *    * AYU_setNumThreads
 *    And these global constants:
 *    * AYU_buf_size
 *    * AYU_string_buf_size
 *  COPYRIGHT
 *    (C) 2010-2013 HLRS, University of Stuttgart
 *    Ayudame is published under the terms of the BSD license.
******/

#ifdef SWIG
%{
#include "Ayudame.h"
%}
// SWIG does not understand __attribute__, so unset it.
#define __attribute__(x)
#endif

#ifndef Ayudame_H
#define Ayudame_H

#define AYU_API_VERSION 1

#include <stdint.h>
#include <unistd.h>

#ifndef AYU_MASTER_TASKID
#  define AYU_MASTER_TASKID 0
#endif

#ifdef __cplusplus
extern "C" {
#endif

/****t* Ayudame/ayu_event_t
 *  NAME
 *    ayu_event_t - enum of events
 *  DESCRIPTION
 *    ayu_event_t consists of the following events:
 *    *  AYU_EVENT_NULL -- "no event", used for initialisation etc.
 *    *  AYU_PREINIT -- establish socket connection. This should be the
 *       first event to be issued. It will notify Ayudame of the Runtime type.
 *    *  AYU_INIT -- can be issued after runtime initialisation. Can contain
 *       the number of processing resources. Temanejo will pause the program
 *       when AYU_INIT is issued by default.
 *    *  AYU_FINISH -- notify of the end of the parallel  execution.
 *    *  AYU_REGISTERFUNCTION -- can be issued to pass the function name of a
 *       taskified function.
 *    *  AYU_ADDTASK -- to notify of the creation of a task instance.
 *    *  AYU_ADDHIDDENTASK -- deprecated
 *    *  AYU_ADDDEPENDENCY -- notify of a dependecy between two tasks. Can
 *       contain two ids (e.g. memory addresses) to dispay coloures edges in
 *       Temanejo.
 *    *  AYU_ADDTASKTOQUEUE -- notify of queueing a task.
 *    *  AYU_PRESELECTTASK -- deprecated
 *    *  AYU_PRERUNTASK -- notify that a task is dequeued and assigned to
 *       a precessing resource.
 *    *  AYU_RUNTASK -- to be issued just before a task is executed. For time
 *       measurement.
 *    *  AYU_POSTRUNTASK -- to be issued right after task finished. For time
 *       measurement.
 *    *  AYU_RUNTASKFAILED -- notify of failure to run task
 *    *  AYU_REMOVETASK -- notify of successfully finished task. Removed from
 *       the graph.
 *    *  AYU_WAITON -- notify of "wait-on" event
 *    *  AYU_BARRIER -- notify of "barrier" event
 *    *  AYU_ADDWAITONTASK -- deprecated
 *
 *  SOURCE
 */

typedef enum ayu_event_t {
  AYU_EVENT_NULL = 0,
  AYU_PREINIT = 1,
  AYU_INIT = 2,
  AYU_FINISH = 3,
  AYU_REGISTERFUNCTION = 4,
  AYU_ADDTASK = 5,
  AYU_ADDHIDDENTASK = 6,
  AYU_ADDDEPENDENCY = 7,
  AYU_ADDTASKTOQUEUE = 8,
  AYU_PRESELECTTASK = 9,
  AYU_PRERUNTASK = 10,
  AYU_RUNTASK = 11,
  AYU_POSTRUNTASK = 12,
  AYU_RUNTASKFAILED = 13,
  AYU_REMOVETASK = 14,
  AYU_WAITON = 15,
  AYU_BARRIER = 16,
  AYU_ADDWAITONTASK = 17
} ayu_event_t;

/******/

/****t* Ayudame_types.h/ayu_request_t
 *  NAME
 *    ayu_request_t - enum of requests
 *  DESCRIPTION
 *    ayu_event_t consists of the following events:
 *    * AYU_REQUEST_NULL -- "no request", used for initialisation
 *    * AYU_NOREQUEST -- explicit "no request"
 *    * AYU_PAUSEONEVENT -- pause request, event upon which the program should
 *      pause is given as third parameter, fourth parameter is 0 for "off"
 *      ("un-pause") and 1 for "on"
 *    * AYU_PAUSEONTASK --  pause request, task id is given as third parameter,
 *      fourth parameter is 0 for "off" ("un-pause") and 1 for "on"
 *    * AYU_PAUSEONFUNCTION -- pause request, function id is given as third
 *      parameter, fourth parameter is 0 for "off" ("un-pause") and 1 for "on"
 *    * AYU_STEP -- run until next pause condition is reached
 *    * AYU_BREAKPOINT -- set breakpoint, i.e. don't assign new tasks, third
 *      parameter is 0 for "off" ("un-pause") and 1 for "on"
 *    * AYU_BLOCKTSK -- to block a specific task. Task id has to be passed.
 *    * AYU_PRIORITISETASK -- to set the priority level of a specific task.
 *    * AYU_SETNUMTHREADS -- to set the number of processing resources.
 *  SOURCE
 */

typedef enum ayu_request_t{
  AYU_REQUEST_NULL = 0,
  AYU_NOREQUEST = 1,
  AYU_PAUSEONEVENT = 2,
  AYU_PAUSEONTASK = 3,
  AYU_PAUSEONFUNCTION = 4,
  AYU_STEP = 5,
  AYU_BREAKPOINT = 6,
  AYU_BLOCKTASK = 7,
  AYU_PRIORITISETASK = 8,
  AYU_SETNUMTHREADS = 9
} ayu_request_t;
/******/

/****t* Ayudame_types.h/ayu_runtime_t
 *  NAME
 *    ayu_runtime_t - enum of runtime ids
 *  DESCRIPTION
 *    ayu_runtime_t consists of the following ids:
 *    * AYU_RT_UNKNOWN -- no or unknown runtime, used for initialisation
 *    * AYU_RT_CPPSS -- CPPSS runtime
 *    * AYU_RT_SMPSS -- SmpSs runtime
 *    * AYU_RT_OMPSS -- OmpSs runtime
 *    * AYU_RT_STARPU -- StarPU runtime
 *    * AYU_RT_FASTFLOW -- FastFlow runtime
 *    * AYU_RT_MPI --  MPI runtime
 *    * AYU_RT_GOMP -- GNU OpenMP runtime
 *    * AYU_RT_CILK -- CILK runtime
 *  SOURCE
 */
typedef enum ayu_runtime_t{
  AYU_RT_UNKNOWN = 0,
  AYU_RT_CPPSS = 1,
  AYU_RT_SMPSS = 2,
  AYU_RT_OMPSS = 3,
  AYU_RT_STARPU = 4,
  AYU_RT_FASTFLOW = 5,
  AYU_RT_MPI = 6,
  AYU_RT_GOMP = 7,
  AYU_RT_CILK = 8
} ayu_runtime_t;
/******/

  /****f* Ayudame/AYU_event
   *  NAME
   *    AYU_event
   *    - callback for the *Ss runtime
   *  SYNOPSIS
   *    AYU_event(event, taskId, pointer)
   *  DESCRIPTION
   *    this function is called from the task parallel runtime to communicate
   *    with the Ayudame package. Events of type ayu_event_t, the task ID
   *    of the calling task and a pointer to additional data can be sent.
   *    AYU_event will connect to a client via tcp sockets and pass on
   *    the information.
   *  PARAMETERS
   *    ayu_event_t event      // type of event
   *    const int64_t taskId   // taskId from which the event is sent
   *    void *p                // pointer to further data
   *  RETURN VALUE
   *    void
   *  SOURCE
   */
#ifndef SWIG
  void AYU_event(ayu_event_t event, const int64_t taskId, void *p)
    __attribute__ ((weak));
#endif
  /******/

  /****f* Ayudame/AYU_getBreakpoint
   *  NAME
   *    AYU_getBreakpoint
   *    - return AYU_breakpoint
   *  SYNOPSIS
   *    breakpoint = AYU_getBreakpoint()
   *  DESCRIPTION
   *    Locks the global variable AYU_breakpoint, returns its value
   *    and unlocks it
   *  PARAMETERS
   *    void
   *  RETURN VALUE
   *    1 if AYU_breakpoint is set
   *    0 if AYU_breakpoint is not set
   *  SOURCE
   */
#ifndef SWIG
  int AYU_getBreakpoint(void) __attribute__ ((weak));
#endif
  /******/

  /****f* Ayudame/AYU_isBlocked
   *  NAME
   *    AYU_isBlocked
   *    - return true if task is blocked
   *  SYNOPSIS
   *    if(AYU_isBlocked(taskId)) { do_something(); }
   *  DESCRIPTION
   *    Locks the global variable AYU_blockedTask using the
   *    breakpoint mutex lock, compares to it
   *    taskId given as parameter and returns result
   *  PARAMETERS
   *    * const int taskId
   *  RETURN VALUE
   *    true if task is blocked
   *    false if task is not blocked
   *  SOURCE
   */
#ifndef SWIG
  unsigned AYU_isBlocked(const int taskId) __attribute__ ((weak));
#endif
  /******/

  /****f* Ayudame/AYU_getPriorityLevel
   *  NAME
   *    AYU_getPriorityLevel
   *    - returns priority of task
   *  SYNOPSIS
   *    AYU_getPriorityLevel(taskId)
   *  DESCRIPTION
   *    returns the priority of a task identified by its id.
   *    If the priority has not been changed in Ayudame it
   *    returns -1.
   *  PARAMETERS
   *    int taskId
   *  RETURN VALUE
   *    -1: if priority has not changed
   *    0:  low (no) priority
   *    positive integer: priority level
   *  SOURCE
   */
#ifndef SWIG
  int AYU_getPriorityLevel(int taskId) __attribute__ ((weak));
#endif
  /******/

  /****f* Ayudame/AYU_getNumThreads
   *  NAME
   *    AYU_getNumThreads
   *    - returns number of threads set by Ayudame
   *  SYNOPSIS
   *    runtime_setNumThreads(AYU_getNumThreads());
   *  DESCRIPTION
   *    returns the number of threads as set by ayudame
   *  PARAMETERS
   *    void
   *  RETURN VALUE
   *    unsigned int: number of threads
   *  SOURCE
   */
#ifndef SWIG
  unsigned int AYU_getNumThreads() __attribute__ ((weak));
#endif
  /******/

  /****f* Ayudame/AYU_setNumThreads
   *  NAME
   *    AYU_setNumThreads
   *    - sets number of threads in the runtime
   *  SYNOPSIS
   *    AYU_setNumThreads(n_threads);
   *  DESCRIPTION
   *    This callback function can be implemented in the StarSs
   *    runtime. If it is not implemented, the client of the
   *    Ayudame library will not be able to change the number of
   *    threads.
   *  PARAMETERS
   *    unsigned int n_threads
   *  RETURN VALUE
   *    void
   */
#ifndef SWIG
  void AYU_setNumThreads(unsigned int n_threads) __attribute__ ((weak));
#endif
  /******/


  /****f* Ayudame/AYU_registerTask
   *  NAME
   *    AYU_registerTask
   *    - obsolete. kept for backward compatibility
   *  SYNOPSIS
   *
   *  DESCRIPTION
   *
   *  PARAMETERS
   *
   *  RETURN VALUE
   *
   *  SOURCE
   */
#ifndef SWIG
  void AYU_registerTask(void*) __attribute__ ((weak));
#endif
  /******/

#ifdef __cplusplus
}
#endif

#endif //Ayudame_H
